/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
+                                                                             +
+	IRQL.c                                                                    +
+                                                                             +
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

DWORD LockAcquired;
DWORD nCPUsLocked;

KIRQL RaiseIRQL()
{
	KIRQL curr;
	KIRQL prev;
	curr = KeGetCurrentIrql();
	prev = curr;
	if(curr < DISPATCH_LEVEL)
	{
		KeRaiseIrql(DISPATCH_LEVEL,&prev);
	}
	return(prev);
}/*end RaiseIRQL()-----------------------------------------------------------*/

void lockRoutine
(
	IN PKDPC dpc,
	IN PVOID context,
	IN PVOID arg1,
	IN PVOID arg2
)
{
	InterlockedIncrement(&nCPUsLocked);
	while(InterlockedCompareExchange(&LockAcquired,1,1)==0)
	{
		__asm
		{ 
			nop; 
		}
	}
	InterlockedDecrement(&nCPUsLocked);
	return;
}/*end lockRoutine()---------------------------------------------------------*/

PKDPC AcquireLock()
{
	PKDPC dpcArray;
	DWORD cpuID;
	DWORD i;
	DWORD nOtherCPUs;

	if(KeGetCurrentIrql()!=DISPATCH_LEVEL){ return(NULL); }
	InterlockedAnd(&LockAcquired,0);
	InterlockedAnd(&nCPUsLocked,0);
	dpcArray = (PKDPC)ExAllocatePool
	(
		NonPagedPool,
		KeNumberProcessors * sizeof(KDPC)
	);
	if(dpcArray==NULL){ return(NULL); }
	cpuID = KeGetCurrentProcessorNumber();
	for(i=0;i<(DWORD)KeNumberProcessors;i++)
	{
		PKDPC dpcPtr = &(dpcArray[i]);
		if(i!=cpuID)
		{
			KeInitializeDpc(dpcPtr,lockRoutine,NULL);
			KeSetTargetProcessorDpc(dpcPtr,(CCHAR)i);
			KeInsertQueueDpc(dpcPtr,NULL,NULL);
		}
	}
	nOtherCPUs = KeNumberProcessors-1;
	InterlockedCompareExchange(&nCPUsLocked, nOtherCPUs, nOtherCPUs);
	while(nCPUsLocked != nOtherCPUs)
	{
		__asm
		{ 
			nop; 
		}
		InterlockedCompareExchange(&nCPUsLocked, nOtherCPUs, nOtherCPUs);
	}
	return(dpcArray);
}/*end AcquireLock()---------------------------------------------------------*/

NTSTATUS ReleaseLock(PVOID dpcPtr)
{
	InterlockedIncrement(&LockAcquired);
	InterlockedCompareExchange(&nCPUsLocked,0,0);
	while(nCPUsLocked != 0)
	{
		__asm
		{ 
			nop; 
		}
		InterlockedCompareExchange(&nCPUsLocked,0,0);
	}
	if(dpcPtr!=NULL)
	{
		ExFreePool(dpcPtr);
	}
	return(STATUS_SUCCESS);
}/*end ReleaseLock()---------------------------------------------------------*/

void LowerIRQL(KIRQL prev)
{
	KeLowerIrql(prev);
	return;
}/*end LowerIRQL()-----------------------------------------------------------*/